/*
   Linux network driver for National Instruments EtherCAT.

   Copyright (C) 2014 National Instruments. All rights reserved.

   This is a driver for Intel(R) I210 Gigabit Ethernet device.
   This driver is mainly used in Linux 3.10 x64 and above and it does _not_
   support Linux 2.6 and 2.4 kernel due to certain new features is introduced in
   this driver.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "niecati210.h"

#include <linux/module.h>
#include <linux/init.h>
#include <linux/pagemap.h>
#include <linux/tcp.h>
#include <linux/kernel.h>
#include <linux/kthread.h>
#include <linux/etherdevice.h>
#include <linux/mm.h>
#include <linux/phy.h>
#include <linux/dma-mapping.h>
#include <linux/io.h>
#include <linux/version.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/ioctl.h>
#include <linux/errno.h>
#include <asm/mman.h>
#include <linux/ktime.h>
#include <linux/if_arp.h>

#include <linux/bitops.h>
#include <linux/vmalloc.h>
#include <linux/netdevice.h>
#include <linux/slab.h>
#include <linux/if.h>
#include <linux/pci.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 4, 0)
#include <linux/pci-aspm.h>
#endif

#include <linux/delay.h>
#include <linux/if_ether.h>
#include <linux/aer.h>

#define EI210_RING_ALIGN  4096
#define EI210_MAX_DEV_NUM 16

static int ei210_dev_num = 0;
static int ei210_major = 0;
static struct class* ei210_class = NULL;

// Following variables are for kernel module parameters
static unsigned int param_dev_num = 0;
static char* param_mac_addr[EI210_MAX_DEV_NUM];
static u8 param_mac_addr_hex[EI210_MAX_DEV_NUM][EI210_ETH_ADDR_LEN];
static unsigned int param_mac_addr_num = 0;

static unsigned int param_master_id_num = 0;
static unsigned int param_master_id[EI210_MAX_DEV_NUM];

static const struct e1000_info *ei210_e1000_info_tbl[] = {
	[board_82575] = &e1000_82575_info,
};

char ei210_driver_name[] = "ni_ethercat_i210k";

static void ei210_reset(struct ei210_adapter *adapter);
static void ei210_irq_disable(struct ei210_adapter *adapter);
static void ei210_release_hw_control(struct ei210_adapter *adapter);
static void ei210_get_hw_control(struct ei210_adapter *adapter);

static void ei210_device_release_driver(const char *mac_addr);
static int ei210_probe(struct pci_dev *pdev,const struct pci_device_id *ent);
static void ei210_remove(struct pci_dev *pdev);

static int ei210_setup_cdev(struct ei210_adapter *adapter);
static void ei210_destroy_cdev(struct ei210_adapter *adapter);

/* User-mode API  */
static int ei210_open_file(struct inode *inode, struct file * file);
static int ei210_close_file(struct inode *inode, struct file *file);
static long ei210_ioctl_file(struct file *file, unsigned int cmd, unsigned long arg);
static int ei210_mmap(struct file *file, struct vm_area_struct *vma);
static ssize_t ei210_no_read(struct file * file, char __user * buf, size_t count, loff_t *pos);
static ssize_t ei210_no_write(struct file * file, const char __user * buf, size_t count, loff_t *pos);
static unsigned int ei210_no_pollfd(struct file *file, poll_table *wait);

static int ei210_setup_rx_resources(struct ei210_transmit_ring *ring);
static int ei210_setup_tx_resources(struct ei210_transmit_ring *ring);

static void ei210_free_tx_resources(struct ei210_transmit_ring *tx_ring);
static void ei210_free_rx_resources(struct ei210_transmit_ring *rx_ring);

static int ei210_alloc_ring(struct ei210_adapter *adapter);
static void ei210_clean_tx_ring(struct ei210_transmit_ring *tx_ring);
static void ei210_clean_rx_ring(struct ei210_transmit_ring *rx_ring);

static void ei210_set_linkspeed(struct ei210_adapter *adapter);
static void ei210_setup_tctl(struct ei210_adapter *adapter);
static void ei210_setup_rctl(struct ei210_adapter *adapter);
static void ei210_setup_mrqc(struct ei210_adapter *adapter);

static void ei210_vm_open(struct vm_area_struct *vma);
static void ei210_vm_close(struct vm_area_struct *vma);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 17, 0)
static vm_fault_t ei210_vm_fault(struct vm_fault *fdata);
#elif defined(NEW_VM_OPERATIONS_STRUCT)
static int ei210_vm_fault(struct vm_fault *fdata);
#else
static int ei210_vm_fault(struct vm_area_struct *area, struct vm_fault *fdata);
#endif

static struct file_operations ei210_fops = {
		.owner   = THIS_MODULE,
		.llseek  = no_llseek,
		.read	= ei210_no_read,
		.write   = ei210_no_write,
		.poll	= ei210_no_pollfd,
		.open	= ei210_open_file,
		.release = ei210_close_file,
		.mmap	= ei210_mmap,
		.unlocked_ioctl = ei210_ioctl_file,
};

static struct vm_operations_struct ei210_mmap_ops = {
		.open   = ei210_vm_open,
		.close  = ei210_vm_close,
		.fault  = ei210_vm_fault
};

// Just to make sure the network device open will NOT crash.
static int ei210_open(struct net_device* ndev)
{
	return -ENODEV;
}

static int ei210_close(struct net_device* ndev)
{
	return 0;
}

static const struct net_device_ops igb_netdev_ops = {
	.ndo_open            = ei210_open,
	.ndo_stop            = ei210_close,
	.ndo_start_xmit      = NULL,
	.ndo_set_rx_mode     = NULL,
	.ndo_set_mac_address = NULL,
	.ndo_do_ioctl        = NULL,
	.ndo_change_mtu      = NULL,
	.ndo_get_stats64     = NULL,
};

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0))
   #ifndef DEFINE_PCI_DEVICE_TABLE
      #define DEFINE_PCI_DEVICE_TABLE(_table) const struct pci_device_id _table[]
   #endif
#endif // For linux kernel version >= 4.8.0

// NOTE only support I210 copper in this EtherCAT driver
static DEFINE_PCI_DEVICE_TABLE(ei210_pci_tbl) = {
	{ PCI_VDEVICE(INTEL, E1000_DEV_ID_I210_COPPER), board_82575 },

	/* required last entry */
	{0, }
};

static struct pci_driver ei210_driver = {
	.name     = ei210_driver_name,
	.id_table = ei210_pci_tbl,
	.probe    = ei210_probe,
	.remove   = ei210_remove,
};

static void ei210_vm_open(struct vm_area_struct *vma)
{
}

static void ei210_vm_close(struct vm_area_struct *vma)
{
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 17, 0)
static vm_fault_t ei210_vm_fault(struct vm_fault *fdata)
#elif defined(NEW_VM_OPERATIONS_STRUCT)
static int ei210_vm_fault(struct vm_fault *fdata)
#else
static int ei210_vm_fault(struct vm_area_struct *area, struct vm_fault *fdata)
#endif
{
	return VM_FAULT_SIGBUS;
}

static void ei210_free_userpage(
	struct ei210_adapter *adapter,
	struct ei210_user_page *userpage)
{
	dma_unmap_page(
		pci_dev_to_dev(adapter->pdev),
		userpage->page_dma,
		PAGE_SIZE,
		DMA_FROM_DEVICE);

	put_page(userpage->page);

	/* take the page out of our list and free it */
	if (userpage->prev)
	{
		userpage->prev->next = userpage->next;
	}

	if (userpage->next)
	{
		userpage->next->prev = userpage->prev;
	}

	if (userpage == adapter->userpages)
	{
		adapter->userpages = userpage->next;
	}
	vfree(userpage);
}

static void ei210_free_all_userpages(struct ei210_adapter *adapter)
{
	struct ei210_user_page *userpage;

	for (userpage = adapter->userpages; userpage; userpage = adapter->userpages)
		ei210_free_userpage(adapter, userpage);
}

static int ei210_setup_cdev(struct ei210_adapter *adapter)
{
	int result = 0;
	int i = 0;
	dev_t dev_num;

	// Find out the master ID by current device mac address.
	for (i = 0; i < param_dev_num; ++i)
	{
		if(memcmp(adapter->hw.mac.addr, param_mac_addr_hex[i], EI210_ETH_ADDR_LEN) == 0)
		{
			break;
		}
	}

	if (i == param_dev_num)
	{
		// We're not supposed to be here!
		// If we're here, that means this driver probed an unexpected device.
		// This guarantees EI210_MAX_DEV_NUM devices could be probed successfully at most.
		printk(KERN_ERR "NI EtherCAT ei210: An unexpected device is found! \n");
		return -ENODEV;
	}

	dev_num = MKDEV(ei210_major, ei210_dev_num);
	cdev_init(&adapter->char_dev, &ei210_fops);

	adapter->char_dev.owner = THIS_MODULE;
	adapter->char_dev.ops = &ei210_fops;
	result = cdev_add(&adapter->char_dev, dev_num, 1);
	if (result)
	{
		printk(KERN_ERR "NI EtherCAT ei210: add character device failed \n");
		return result;
	}

	device_create(ei210_class, NULL, dev_num, NULL, kNIECATI210DevNameFmt, param_master_id[i]);
	return 0;
}

static void ei210_destroy_cdev(struct ei210_adapter *adapter)
{
	ei210_free_tx_resources(adapter->tx_ring);
	ei210_free_rx_resources(adapter->rx_ring);

	ei210_free_all_userpages(adapter);

	device_destroy(ei210_class, adapter->char_dev.dev);
	cdev_del(&adapter->char_dev);
}

static void ei210_device_release_driver(const char *mac_addr)
{
	struct net_device *netdev;

	rtnl_lock();
	netdev = dev_getbyhwaddr_rcu(&init_net, ARPHRD_ETHER, mac_addr);
	rtnl_unlock();

	if (netdev)
		device_release_driver(netdev->dev.parent);
}

static int ei210_alloc_ring(struct ei210_adapter *adapter)
{
	struct ei210_transmit_ring *ring;
	size_t size;

	size = sizeof(struct ei210_transmit_ring);
	ring = kzalloc(size, GFP_KERNEL);
	if (!ring)
		return -ENOMEM;

	ring->dev = &adapter->pdev->dev;
	ring->count = EI210_DEFAULT_TXD;
	adapter->tx_ring = ring;

	ring = kzalloc(size, GFP_KERNEL);
	if (!ring)
		return -ENOMEM;

	ring->dev = &adapter->pdev->dev;
	ring->count = EI210_DEFAULT_RXD;
	adapter->rx_ring = ring;

	return 0;
}

static int ei210_sw_init(struct ei210_adapter *adapter)
{
	struct e1000_hw *hw = &adapter->hw;
	struct net_device *netdev = adapter->netdev;
	struct pci_dev *pdev = adapter->pdev;

	pci_read_config_word(pdev, PCI_COMMAND, &hw->bus.pci_cmd_word);

	/* Don't add VLAN_HLEN because ECAT does not support VLAN.*/
	adapter->max_frame_size = netdev->mtu + ETH_HLEN + ETH_FCS_LEN;
	adapter->min_frame_size = ETH_ZLEN + ETH_FCS_LEN;

	/* Explicitly disable IRQ since the NIC can be in any state. */
	ei210_irq_disable(adapter);

	if (hw->mac.type >= e1000_i350)
		adapter->flags &= ~EI210_FLAG_DMAC;

	return 0;
}

void igb_power_up_link(struct ei210_adapter *adapter)
{
	ei210_reset_phy(&adapter->hw);

	if (adapter->hw.phy.media_type == e1000_media_type_copper)
		igb_power_up_phy_copper(&adapter->hw);
	else
	{
		// Unsupported media type. Only support I210 copper
		dev_err(&adapter->pdev->dev, "Unsupported phy.media_type %d\n",
			adapter->hw.phy.media_type);
	}
}

static int ei210_probe(struct pci_dev *pdev,const struct pci_device_id *ent)
{
	struct net_device *netdev;
	struct ei210_adapter *adapter;
	struct e1000_hw *hw;
	const struct e1000_info *ei = ei210_e1000_info_tbl[ent->driver_data];
	unsigned long mmio_start, mmio_len;
	int err;

	/* Catch broken hardware that put the wrong VF device ID in
	 * the PCIe SR-IOV capability.
	 */
	if (pdev->is_virtfn)
	{
		WARN(1, KERN_ERR "%s (%hx:%hx) should not be a VF!\n",
			pci_name(pdev), pdev->vendor, pdev->device);
		return -EINVAL;
	}

	dev_dbg(&pdev->dev, "ei210_probe device vendor=%hx id=%hx revision=%hx\n",
		pdev->vendor,
		pdev->device,
		pdev->revision);

	err = pci_enable_device_mem(pdev);
	if (err)
		return err;

	// DMA bit mask. Try 64 first, then 32
	err = dma_set_mask(&pdev->dev, DMA_BIT_MASK(64));
	if (!err)
	{
		err = dma_set_coherent_mask(&pdev->dev, DMA_BIT_MASK(64));
	}
	else
	{
		err = dma_set_mask(&pdev->dev, DMA_BIT_MASK(32));
		if (!err)
		{
			err = dma_set_coherent_mask(&pdev->dev,
				DMA_BIT_MASK(32));
		}
	}

	if (err)
	{
		dev_err(&pdev->dev, "No usable DMA configuration, aborting\n");
		goto err_dma;
	}

	err = pci_request_selected_regions(
		pdev,
		pci_select_bars(pdev,
		IORESOURCE_MEM),
		ei210_driver_name);
	if (err)
		goto err_pci_reg;

	pci_enable_pcie_error_reporting(pdev);

	pci_set_master(pdev);
	pci_save_state(pdev);

	netdev = alloc_etherdev_mq(sizeof(struct ei210_adapter), 1);
	if (!netdev)
	{
		err = -ENOMEM;
		goto err_alloc_etherdev;
	}

	SET_NETDEV_DEV(netdev, &pdev->dev);

	pci_set_drvdata(pdev, netdev);
	adapter = netdev_priv(netdev);
	adapter->netdev = netdev;
	adapter->pdev = pdev;
	hw = &adapter->hw;
	hw->back = adapter;

	mmio_start = pci_resource_start(pdev, 0);
	mmio_len = pci_resource_len(pdev, 0);

	hw->hw_addr = ioremap(mmio_start, mmio_len);
	if (!hw->hw_addr)
	{
		err = -EIO;
		goto err_ioremap;
	}

	netdev->netdev_ops = &igb_netdev_ops;

	/* PCI config space info */
	hw->vendor_id = pdev->vendor;
	hw->device_id = pdev->device;
	hw->revision_id = pdev->revision;
	hw->subsystem_vendor_id = pdev->subsystem_vendor;
	hw->subsystem_device_id = pdev->subsystem_device;

	/* Copy the default MAC, PHY and NVM function pointers */
	memcpy(&hw->mac.ops, ei->mac_ops, sizeof(hw->mac.ops));
	memcpy(&hw->phy.ops, ei->phy_ops, sizeof(hw->phy.ops));
	memcpy(&hw->nvm.ops, ei->nvm_ops, sizeof(hw->nvm.ops));

	/* Initialize skew-specific constants */
	err = ei->get_invariants(hw);
	if (err)
		goto err_sw_init;

	/* setup the private structure */
	err = ei210_sw_init(adapter);
	if (err)
	{
		goto err_sw_init;
	}

	igb_get_bus_info_pcie(hw);

	hw->phy.autoneg_wait_to_complete = false;

	/* Copper options */
	if (hw->phy.media_type == e1000_media_type_copper)
	{
		hw->phy.mdix = EI210_AUTO_ALL_MODES;
		hw->phy.disable_polarity_correction = false;
		hw->phy.ms_type = e1000_ms_hw_default;
	}

	if (igb_check_reset_block(hw))
		dev_info(&pdev->dev,
			"PHY reset is blocked due to SOL/IDER session.\n");

	hw->mac.ops.reset_hw(hw);

	if (hw->nvm.ops.validate(hw) < 0)
	{
		dev_err(&pdev->dev, "The NVM Checksum Is Not Valid\n");
		err = -EIO;
		goto err_eeprom;
	}

	/* copy the MAC address out of the NVM */
	if (hw->mac.ops.read_mac_addr(hw))
		dev_err(&pdev->dev, "NVM Read Error\n");

	memcpy(adapter->netdev->dev_addr, hw->mac.addr, ETH_ALEN);

	hw->mac.autoneg = true;
	hw->phy.autoneg_advertised = 0x2f;

	hw->fc.requested_mode = e1000_fc_default;
	hw->fc.current_mode = e1000_fc_default;

	/* Not validate MDI setting here because all MDI settings are supported on 82580 and newer. */

	/* Do not support wake on port */
	device_set_wakeup_enable(&adapter->pdev->dev, false);

	/* reset the hardware with the new settings */
	ei210_reset(adapter);

	/* let the f/w know that the h/w is now under the control of the
	 * driver. */
	ei210_get_hw_control(adapter);

	strcpy(netdev->name, "eth%d");
	err = register_netdev(netdev);
	if (err)
		goto err_register;

	netif_carrier_off(netdev);

	/* Disable EEE for EtherCAT*/
	hw->dev_spec._82575.eee_disable = true;
	igb_set_eee_i350(hw);

	err = ei210_alloc_ring(adapter);
	if (err)
		goto err_register;

	err = ei210_setup_cdev(adapter);
	if (err)
		goto err_register;

	igb_power_up_link(adapter);

	++ei210_dev_num;
	return 0;

err_register:
	ei210_release_hw_control(adapter);
err_eeprom:
	if (!igb_check_reset_block(hw))
		ei210_reset_phy(hw);

	if (hw->flash_address)
		iounmap(hw->flash_address);
err_sw_init:
	iounmap(hw->hw_addr);
err_ioremap:
	free_netdev(netdev);
err_alloc_etherdev:
	pci_release_selected_regions(pdev, pci_select_bars(pdev, IORESOURCE_MEM));
err_pci_reg:
err_dma:
	pci_disable_device(pdev);
	return err;
}

static void ei210_irq_disable(struct ei210_adapter *adapter)
{
	struct e1000_hw *hw = &adapter->hw;

	wr32(E1000_IAM, 0);
	wr32(E1000_IMC, ~0);
	wrfl();
}

static void ei210_release_hw_control(struct ei210_adapter *adapter)
{
	struct e1000_hw *hw = &adapter->hw;
	u32 ctrl_ext;

	/* Let firmware take over control of h/w */
	ctrl_ext = rd32(E1000_CTRL_EXT);

	wr32(E1000_CTRL_EXT, ctrl_ext & ~E1000_CTRL_EXT_DRV_LOAD);
}

static void ei210_get_hw_control(struct ei210_adapter *adapter)
{
	struct e1000_hw *hw = &adapter->hw;
	u32 ctrl_ext;

	/* Let firmware know the driver has taken over */
	ctrl_ext = rd32(E1000_CTRL_EXT);
	wr32(E1000_CTRL_EXT, ctrl_ext | E1000_CTRL_EXT_DRV_LOAD);
}

static ssize_t ei210_no_read(struct file * file, char __user * buf, size_t count, loff_t *pos)
{
	return -EINVAL;
}

static ssize_t ei210_no_write(struct file * file, const char __user * buf, size_t count, loff_t *pos)
{
	return -EINVAL;
}

static unsigned int ei210_no_pollfd(struct file *file, poll_table *wait)
{
	return -EINVAL;
}

static int ei210_open_file(struct inode *inode, struct file * file)
{
	int err = 0;
	struct ei210_adapter *adapter = container_of(
		inode->i_cdev,
		struct ei210_adapter,
		char_dev);

	file->private_data = adapter;

	/* IRQ is already disabled in sw_init */

	err = ei210_setup_tx_resources(adapter->tx_ring);
	if (err)
	{
		printk(KERN_ERR "ei210_open_file: set tx reousrces fail\n");
		return err;
	}

	/* allocate receive descriptors */
	err = ei210_setup_rx_resources(adapter->rx_ring);
	if (err)
	{
		printk(KERN_ERR "ei210_open_file: set rx reousrces fail\n");
		goto err_setup_misc_rx;
	}

	ei210_set_linkspeed(adapter);
	ei210_setup_tctl(adapter);
	ei210_setup_mrqc(adapter);
	ei210_setup_rctl(adapter);

	return 0;

err_setup_misc_rx:
	ei210_free_tx_resources(adapter->tx_ring);

	return err;
}

static long ei210_map_ring(struct file *file, void __user *arg)
{
	struct ei210_adapter *adapter;
	struct ei210_buf_cmd req;
	struct ei210_transmit_ring* ring;

	if (copy_from_user(&req, arg, sizeof(req)))
		return -EFAULT;

	adapter = file->private_data;
	if (!adapter)
		return -ENOENT;

	/* Req flags, Tx: 0, Rx: 1 */
	ring = req.flags ? adapter->rx_ring : adapter->tx_ring;

	if (ring->mapped)
		return -EINVAL;

	ring->mapped = true;

	req.physaddr = ring->dma;
	req.mmap_size = ring->size;

	if (copy_to_user(arg, &req, sizeof(req)))
	{
		dev_err(&adapter->pdev->dev,
			"ei210_map_ring: copy_to_user failed\n");

		return -EFAULT;
	}

	return 0;
}

static long ei210_map_buf(struct file *file, void __user *arg)
{
	struct ei210_adapter *adapter;
	struct ei210_buf_cmd req;

	struct page *page = NULL;
	dma_addr_t page_dma;
	struct ei210_user_page *userpage = NULL;

	if (copy_from_user(&req, arg, sizeof(req)))
		return -EFAULT;

	adapter = file->private_data;
	if (!adapter)
		return -ENOENT;

	userpage = vzalloc(sizeof(struct ei210_user_page));
	if (unlikely(!userpage))
		return -ENOMEM;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0)
	page = alloc_page(GFP_ATOMIC);
#else
	page = alloc_page(GFP_ATOMIC | __GFP_COLD);
#endif
	if (unlikely(!page))
	{
		vfree(userpage);
		return -ENOMEM;
	}

	page_dma = dma_map_page(pci_dev_to_dev(adapter->pdev), page,
					0, PAGE_SIZE,
					DMA_FROM_DEVICE);

	if (dma_mapping_error(pci_dev_to_dev(adapter->pdev), page_dma))
	{
		vfree(userpage);
		put_page(page);
		return -ENOMEM;
	}

	/* Not support multi-threaded access here */
	if (adapter->userpages == NULL)
		adapter->userpages = userpage;
	else
	{
		userpage->next = adapter->userpages;
		adapter->userpages->prev = userpage;
		adapter->userpages = userpage;
	}

	adapter->userpages->page = page;
	adapter->userpages->page_dma = page_dma;

	req.physaddr = page_dma;
	req.mmap_size = PAGE_SIZE;

	if (copy_to_user(arg, &req, sizeof(req)))
	{
		/* Here resources are not free because we can free them when close */
		dev_err(&adapter->pdev->dev,
			"ei210_map_userpage: copy_to_user failed\n");

		return -EFAULT;
	}

	return 0;
}

static long ei210_unmap_ring(struct file *file, void __user *arg)
{
	struct ei210_adapter *adapter;
	struct ei210_buf_cmd req;
	struct ei210_transmit_ring* ring;

	if (copy_from_user(&req, arg, sizeof(req)))
		return -EFAULT;

	adapter = file->private_data;
	if (!adapter)
		return -ENOENT;

	/* Req flags, Tx: 0, Rx: 1 */
	ring = req.flags ? adapter->rx_ring : adapter->tx_ring;

	if (!ring->mapped)
		return -EINVAL;

	ring->mapped = false;
	return 0;
}

static long ei210_unmap_buf(struct file *file, void __user *arg)
{
	struct ei210_adapter *adapter;
	struct ei210_buf_cmd req;

	struct ei210_user_page *userpage;

	if (copy_from_user(&req, arg, sizeof(req)))
		return -EFAULT;

	adapter = file->private_data;
	if (!adapter)
		return -ENOENT;

	/* have to find the corresponding page to free */
	for (userpage = adapter->userpages; userpage != NULL; userpage = userpage->next)
	{
		if (req.physaddr == userpage->page_dma)
			break;
	}

	if (userpage == NULL)
		return -EINVAL;

	ei210_free_userpage(adapter, userpage);
	return 0;
}

static int ei210_file_get_mac(struct file *file, void __user *argp)
{
	struct ei210_adapter *adapter;
	struct e1000_hw *hw;
	struct e1000_mac_info *mac;
	int err = 0, i;
	u64 mac_addr;

	if (copy_from_user(&mac_addr, argp, sizeof(mac_addr)))
		return -EFAULT;

	adapter = file->private_data;
	if (!adapter)
		return -ENOENT;

	hw = &adapter->hw;
	mac = &hw->mac;

	mac_addr = 0;

	for (i = 0; i < EI210_ETH_ADDR_LEN; ++i)
	{
		mac_addr |= (u64)(mac->addr[i] & 0xFF) << ((EI210_ETH_ADDR_LEN - 1 - i) * 8);
	}

	if (copy_to_user(argp, &mac_addr, sizeof(mac_addr)))
	{
		dev_err(&adapter->pdev->dev,
			"ei210_file_get_mac: copy_to_user failed\n");

		err = -EFAULT;
	}

	return err;
}

static int ei210_unbind(struct file *file)
{
	struct ei210_adapter *adapter;
	adapter = file->private_data;

	if (NULL == adapter)
		return -EBADFD;

	file->private_data = NULL;
	return 0;
}

static int ei210_file_bind(struct file *file, void __user *argp)
{
	struct ei210_adapter *adapter;
	struct ei210_bind_cmd req;
	int err = 0;

	if (copy_from_user(&req, argp, sizeof(req)))
		return -EFAULT;

	adapter = file->private_data;

	if (NULL == adapter)
		return -ENOENT;

	req.mmap_size = 0;
	req.mmap_size = pci_resource_len(adapter->pdev, 0);

	if (copy_to_user(argp, &req, sizeof(req)))
		err = -EFAULT;

	return err;
}


static long ei210_ioctl_file(struct file *file, unsigned int cmd, unsigned long arg)
{
	void __user * argp = (void __user *) arg;
	int err = 0;

	switch (cmd)
	{
		case EI210_IOC_UNBIND:
			err = ei210_unbind(file);
			break;

		case EI210_IOC_MAPRING:
			err = ei210_map_ring(file, argp);
			break;

		case EI210_IOC_MAPBUF:
			err = ei210_map_buf(file, argp);
			break;

		case EI210_IOC_UNMAPRING:
			err = ei210_unmap_ring(file, argp);
			break;

		case EI210_IOC_UNMAPBUF:
			err = ei210_unmap_buf(file, argp);
			break;

		case EI210_IOC_BIND:
			err = ei210_file_bind(file, argp);
			break;

		case EI210_IOC_MACADDR:
			err = ei210_file_get_mac(file, argp);
			break;

		default:
			err = -EINVAL;
			break;
	};

	return err;
}

static int ei210_close_file(struct inode *inode, struct file *file)
{
	struct ei210_adapter *adapter = file->private_data;
	int err;

	if (!adapter)
		return -ENOENT;

	/* force not mapped because users might forget it*/
	adapter->tx_ring->mapped = false;
	ei210_free_tx_resources(adapter->tx_ring);

	/* force not mapped because users might forget it*/
	adapter->rx_ring->mapped = false;
	ei210_free_rx_resources(adapter->rx_ring);

	ei210_free_all_userpages(adapter);

	err = ei210_unbind(file);
	return err;
}

static int ei210_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct ei210_adapter *adapter = file->private_data;
	unsigned long size  = vma->vm_end - vma->vm_start;
	dma_addr_t pgoff = vma->vm_pgoff;
	dma_addr_t physaddr;

	if (NULL == adapter)
		return -ENODEV;

	if (pgoff == 0)
		physaddr = pci_resource_start(adapter->pdev, 0) >> PAGE_SHIFT;
	else
		physaddr = pgoff;

	vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);

	if (remap_pfn_range(
			vma,
			vma->vm_start,
			physaddr,
			size,
			vma->vm_page_prot))
		return -EAGAIN;

	vma->vm_ops = &ei210_mmap_ops;
	return 0;
}

static int ei210_setup_tx_resources(struct ei210_transmit_ring *tx_ring)
{
	struct device *dev = tx_ring->dev;
	int size;

	size = sizeof(struct ei210_tx_buffer) * tx_ring->count;
	tx_ring->tx_buffer_info = vzalloc(size);
	if (!tx_ring->tx_buffer_info)
		goto err;

	/* round up to nearest 4K */
	tx_ring->size = tx_ring->count * sizeof(union e1000_adv_tx_desc);
	tx_ring->size = ALIGN(tx_ring->size, EI210_RING_ALIGN);

	tx_ring->desc = dma_alloc_coherent(
		dev,
		tx_ring->size,
		&tx_ring->dma,
		GFP_KERNEL);

	if (!tx_ring->desc)
		goto err;

	return 0;

err:
	vfree(tx_ring->tx_buffer_info);
	dev_err(dev,
		"Unable to allocate memory for the transmit descriptor ring\n");
	return -ENOMEM;
}

static int ei210_setup_rx_resources(struct ei210_transmit_ring *rx_ring)
{
	struct device *dev = rx_ring->dev;
	int size, desc_len;

	size = sizeof(struct ei210_rx_buffer) * rx_ring->count;
	rx_ring->rx_buffer_info = vzalloc(size);
	if (!rx_ring->rx_buffer_info)
		goto err;

	desc_len = sizeof(union e1000_adv_rx_desc);

	/* Round up to nearest 4K */
	rx_ring->size = rx_ring->count * desc_len;
	rx_ring->size = ALIGN(rx_ring->size, EI210_RING_ALIGN);

	rx_ring->desc = dma_alloc_coherent(
		dev,
		rx_ring->size,
		&rx_ring->dma,
		GFP_KERNEL);

	if (!rx_ring->desc)
		goto err;

	return 0;

err:
	vfree(rx_ring->rx_buffer_info);
	rx_ring->rx_buffer_info = NULL;
	dev_err(dev, "Unable to allocate memory for the receive descriptor ring\n");
	return -ENOMEM;
}

/**
 * ei210_free_tx_resources - Free Tx Resources per Queue
 * @tx_ring: Tx descriptor ring for a specific queue
 *
 * Free all transmit software resources
 **/
static void ei210_free_tx_resources(struct ei210_transmit_ring *tx_ring)
{
	ei210_clean_tx_ring(tx_ring);

	vfree(tx_ring->tx_buffer_info);
	tx_ring->tx_buffer_info = NULL;

	/* if not set, then don't free */
	if (!tx_ring->desc)
		return;

	dma_free_coherent(
		tx_ring->dev,
		tx_ring->size,
		tx_ring->desc,
		tx_ring->dma);

	tx_ring->desc = NULL;
}

static void ei210_unmap_and_free_tx_resource(
	struct ei210_transmit_ring *ring,
	struct ei210_tx_buffer *tx_buffer)
{
	if (dma_unmap_len(tx_buffer, len))
		dma_unmap_page(ring->dev,
			dma_unmap_addr(tx_buffer, dma),
			dma_unmap_len(tx_buffer, len),
			DMA_TO_DEVICE);

	tx_buffer->next_to_watch = NULL;
	dma_unmap_len_set(tx_buffer, len, 0);
}

/**
 * ei210_clean_tx_ring - Free Tx Buffers
 * @tx_ring: ring to be cleaned
 **/
static void ei210_clean_tx_ring(struct ei210_transmit_ring *tx_ring)
{
	struct ei210_tx_buffer *buffer_info;
	unsigned long size;
	u16 i;

	if (!tx_ring->tx_buffer_info)
		return;

	for (i = 0; i < tx_ring->count; i++)
	{
		buffer_info = &tx_ring->tx_buffer_info[i];
		ei210_unmap_and_free_tx_resource(tx_ring, buffer_info);
	}

	size = sizeof(struct ei210_tx_buffer) * tx_ring->count;
	memset(tx_ring->tx_buffer_info, 0, size);

	/* Zero out the descriptor ring */
	memset(tx_ring->desc, 0, tx_ring->size);
}

/**
 * ei210_free_rx_resources - Free Rx Resources
 * @rx_ring: ring to clean the resources from
 *
 * Free all receive software resources
 **/
static void ei210_free_rx_resources(struct ei210_transmit_ring *rx_ring)
{
	ei210_clean_rx_ring(rx_ring);

	vfree(rx_ring->rx_buffer_info);
	rx_ring->rx_buffer_info = NULL;

	/* if not set, then don't free */
	if (!rx_ring->desc)
		return;

	dma_free_coherent(
		rx_ring->dev,
		rx_ring->size,
		rx_ring->desc,
		rx_ring->dma);

	rx_ring->desc = NULL;
}

/**
 * ei210_clean_rx_ring - Free Rx Buffers per Queue
 * @rx_ring: ring to free buffers from
 **/
static void ei210_clean_rx_ring(struct ei210_transmit_ring *rx_ring)
{
	unsigned long size;
	u16 i;

	if (!rx_ring->rx_buffer_info)
		return;

	for (i = 0; i < rx_ring->count; i++)
	{
		struct ei210_rx_buffer *buffer_info = &rx_ring->rx_buffer_info[i];

		if (!buffer_info->page)
			continue;

		dma_unmap_page(rx_ring->dev,
			buffer_info->dma,
			PAGE_SIZE,
			DMA_FROM_DEVICE);

		__free_page(buffer_info->page);

		buffer_info->page = NULL;
	}

	size = sizeof(struct ei210_rx_buffer) * rx_ring->count;
	memset(rx_ring->rx_buffer_info, 0, size);

	/* Zero out the descriptor ring */
	memset(rx_ring->desc, 0, rx_ring->size);
}

static void ei210_set_linkspeed(struct ei210_adapter *adapter)
{
	struct e1000_hw *hw = &adapter->hw;
	u32 ctrl;

	ctrl = ei210_read_reg(hw, E1000_CTRL);
	ctrl |= E1000_CTRL_FRCDPX | E1000_CTRL_FRCSPD | E1000_CTRL_SPD_100;

	ei210_write_reg(hw, E1000_TCTL, ctrl);
}

void ei210_setup_tctl(struct ei210_adapter *adapter)
{
	struct e1000_hw *hw = &adapter->hw;
	u32 tctl;

	/* disable queue 0 which is enabled by default on 82575 and 82576 */
	ei210_write_reg(hw, E1000_TXDCTL(0), 0);

	/* Program the Transmit Control Register */
	tctl = ei210_read_reg(hw, E1000_TCTL);
	tctl &= ~E1000_TCTL_CT;
	tctl |= E1000_TCTL_PSP | E1000_TCTL_RTLC |
		(E1000_COLLISION_THRESHOLD << E1000_CT_SHIFT);

	/* Enable transmits */
	tctl |= E1000_TCTL_EN;

	ei210_write_reg(hw, E1000_TCTL, tctl);
}

void ei210_setup_mrqc(struct ei210_adapter *adapter)
{
	struct e1000_hw *hw = &adapter->hw;
	u32 mrqc = 0;

	ei210_write_reg(hw, E1000_MRQC, mrqc);
}

void ei210_setup_rctl(struct ei210_adapter *adapter)
{
	struct e1000_hw *hw = &adapter->hw;
	u32 rctl;

	rctl = ei210_read_reg(hw, E1000_RCTL);

	rctl &= ~(3 << E1000_RCTL_MO_SHIFT);
	rctl &= ~(E1000_RCTL_LBM_TCVR | E1000_RCTL_LBM_MAC);

	rctl |= E1000_RCTL_EN | E1000_RCTL_BAM | E1000_RCTL_RDMTS_HALF |
		(hw->mac.mc_filter_type << E1000_RCTL_MO_SHIFT);
	/*
	 * enable stripping of CRC. It's unlikely this will break BMC
	 * redirection as it did with e1000. Newer features require
	 * that the HW strips the CRC.
	 */
	rctl |= E1000_RCTL_SECRC;

	/* disable store bad packets and clear size bits. */
	rctl &= ~(E1000_RCTL_SBP | E1000_RCTL_SZ_256);

	/* enable LPE to prevent packets larger than max_frame_size */
	rctl |= E1000_RCTL_LPE;

	/* disable queue 0 to prevent tail write w/o re-config */
	ei210_write_reg(hw, E1000_RXDCTL(0), 0);

	ei210_write_reg(hw, E1000_RCTL, rctl);
}

void ei210_reset(struct ei210_adapter *adapter)
{
	struct pci_dev *pdev = adapter->pdev;
	struct e1000_hw *hw = &adapter->hw;

	/* Allow time for pending master requests to run */
	hw->mac.ops.reset_hw(hw);
	wr32(E1000_WUC, 0);

	if (hw->mac.ops.init_hw(hw))
		dev_err(&pdev->dev, "Hardware Error\n");

	/* Flow control settings reset on hardware reset, so guarantee flow
	 * control is off when forcing speed.
	 */
	if (!hw->mac.autoneg)
		igb_force_mac_fc(hw);

	/* Disable DMA Coalescing for EtherCAT */
	/* Do nothing for I210 because it is disabled */

	/* EtherCAT does not need VLAN feature (ETHERNET_IEEE_VLAN_TYPE) */

	ei210_get_phy_info(hw);
}

s32 igb_read_pcie_cap_reg(struct e1000_hw *hw, u32 reg, u16 *value)
{
	struct ei210_adapter *adapter = hw->back;

	if (pcie_capability_read_word(adapter->pdev, reg, value))
		return -E1000_ERR_CONFIG;

	return 0;
}

s32 igb_write_pcie_cap_reg(struct e1000_hw *hw, u32 reg, u16 *value)
{
	struct ei210_adapter *adapter = hw->back;

	if (pcie_capability_write_word(adapter->pdev, reg, *value))
		return -E1000_ERR_CONFIG;

	return 0;
}

s32 igb_read_i2c_byte(
	struct e1000_hw *hw,
	u8 byte_offset,
	u8 dev_addr, u8 *data)
{
	return E1000_ERR_I2C;
}

s32 igb_write_i2c_byte(
	struct e1000_hw *hw,
	u8 byte_offset,
	u8 dev_addr, u8 data)
{
	return E1000_ERR_I2C;
}

static void ei210_remove(struct pci_dev *pdev)
{
	struct net_device *netdev = pci_get_drvdata(pdev);
	struct ei210_adapter *adapter = netdev_priv(netdev);
	struct e1000_hw *hw = &adapter->hw;

	ei210_release_hw_control(adapter);

	unregister_netdev(netdev);

	iounmap(hw->hw_addr);

	pci_release_selected_regions(pdev, pci_select_bars(pdev, IORESOURCE_MEM));

	free_netdev(netdev);

	pci_disable_pcie_error_reporting(pdev);

	pci_disable_device(pdev);

	ei210_destroy_cdev(adapter);
}

static bool macaddr_str2hex(const char* mac_str, unsigned char* mac_hex)
{
	int index = 0;
	int res = 0;
	char mac_str_buf[EI210_ETH_ADDR_STR_LEN];
	char* mac_dup = mac_str_buf;
	char* mac_seg;
	int mac_str_len = strlen(mac_str);

	if (mac_str_len != EI210_ETH_ADDR_STR_LEN -1) // MAC address string length is not correct.
	{
		return false;
	}

	memset(mac_hex, 0, EI210_ETH_ADDR_LEN);
	memset(mac_dup, 0, EI210_ETH_ADDR_STR_LEN);
	memcpy(mac_dup, mac_str, mac_str_len);

	for (index = 0; index < EI210_ETH_ADDR_LEN; ++index)
	{
		mac_seg = strsep(&mac_dup, ":");
		if (mac_seg == NULL)
			break;

		res = kstrtou8(mac_seg, 16, &mac_hex[index]);
		if (res != 0)
			break;
	}

	return (index == EI210_ETH_ADDR_LEN);
}

static int __init ei210_init(void)
{
	int err = 0;
	int index = 0;
	dev_t dev_num;

	param_dev_num = param_mac_addr_num;
	if (param_mac_addr_num != param_master_id_num)
	{
		param_dev_num = param_mac_addr_num < param_master_id_num ? param_mac_addr_num : param_master_id_num;
		printk(KERN_INFO "NI EtherCAT ei210: param_mac_addr_num doesn't match with param_master_id_num."
			"param_mac_addr_num = %u, param_master_id_num = %u \n", param_mac_addr_num, param_master_id_num);
	}

	err = alloc_chrdev_region(&dev_num, 0, param_dev_num, kNIECATI210ModuleName);
	if (err < 0)
	{
		printk(KERN_ERR "NI EtherCAT ei210: register character device failed \n");
		return err;
	}

	ei210_major = MAJOR(dev_num);

	ei210_class = class_create(THIS_MODULE, kNIECATI210ClassName);
	if (IS_ERR(ei210_class))
	{
		printk(KERN_ERR "NI EtherCAT ei210: create device class failed \n");
		return PTR_ERR(ei210_class);
	}

	// Detach specified devices from original driver.
	for (index = 0; index < param_dev_num; ++index)
	{
		if (macaddr_str2hex(param_mac_addr[index], param_mac_addr_hex[index]))
		{
			ei210_device_release_driver(param_mac_addr_hex[index]);
		}
	}

	err = pci_register_driver(&ei210_driver);

	return err;
}

static void __exit ei210_exit(void)
{
	pci_unregister_driver(&ei210_driver);

	class_destroy(ei210_class);
	unregister_chrdev_region(MKDEV(ei210_major, 0), param_dev_num);
}

module_param_array(param_mac_addr, charp, &param_mac_addr_num, S_IRUGO);
module_param_array(param_master_id, uint, &param_master_id_num, S_IRUGO);

module_init(ei210_init);
module_exit(ei210_exit);

MODULE_AUTHOR("National Instruments");
MODULE_DESCRIPTION("National Instruments EtherCAT network driver for I210");
MODULE_LICENSE("GPL v2");
