/*
   Linux network driver for National Instruments EtherCAT.

   Copyright (C) 2014 National Instruments. All rights reserved.

   This is a driver for Intel(R) I210 Gigabit Ethernet device.
   This driver is mainly used in Linux 3.10 x64 and above and it does _not_
   support Linux 2.6 and 2.4 kernel due to certain new features is introduced in
   this driver.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _NIECATI210_H_
#define _NIECATI210_H_

#include <linux/types.h>
#include <asm/page.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#include <linux/cdev.h>

#include "e1000_mac.h"
#include "e1000_82575.h"

/* IOCTL code, constants and structures */
#define EI210_IOC_BIND       _IOW('E', 200, int)
#define EI210_IOC_UNBIND     _IOW('E', 201, int)
#define EI210_IOC_MAPRING    _IOW('E', 202, int)
#define EI210_IOC_UNMAPRING  _IOW('E', 203, int)
#define EI210_IOC_MAPBUF     _IOW('E', 204, int)
#define EI210_IOC_UNMAPBUF   _IOW('E', 205, int)
#define EI210_IOC_MACADDR    _IOW('E', 206, int)

#define EI210_BIND_NAMESZ  24
#define EI210_ETH_ADDR_LEN 6
#define EI210_ETH_ADDR_STR_LEN (EI210_ETH_ADDR_LEN * 3)

static const char* const kNIECATI210ModuleName = "niecati210k";
static const char* const kNIECATI210ClassName = "ethercat";
static const char* const kNIECATI210DevNameFmt = "niecat%u";

struct ei210_bind_cmd {
	char	iface[EI210_BIND_NAMESZ];
	u32	mmap_size;
};

/* used with both map/unmap ring & buf ioctls */
struct ei210_buf_cmd {
	u64 physaddr;
	u32 reseved;
	u32 mmap_size;
	u32 flags;
};

/* Internal defines and structures */
#define EI210_FLAG_DMAC       (1 << 4)

#define EI210_DEFAULT_TXD     256
#define EI210_DEFAULT_RXD     256
#define EI210_AUTO_ALL_MODES  0

struct ei210_user_page {
	struct ei210_user_page *prev;
	struct ei210_user_page *next;
	struct page *page;
	dma_addr_t page_dma;
};

struct ei210_transmit_ring {
	struct device *dev;		/* device pointer for dma mapping */

	union {				/* array of buffer info structs */
		struct ei210_tx_buffer *tx_buffer_info;
		struct ei210_rx_buffer *rx_buffer_info;
	};

	void *desc;			/* descriptor ring memory */

	dma_addr_t dma;			/* phys address of the ring */
	unsigned int  size;		/* length of desc. ring in bytes */

	u16 count;			/* number of desc. in the ring */

	bool mapped;
} ____cacheline_internodealigned_in_smp;

/* board specific private data structure */
struct ei210_adapter {
	struct net_device *netdev;
	unsigned int flags;

	struct ei210_transmit_ring *tx_ring;
	struct ei210_transmit_ring *rx_ring;
	u32 max_frame_size;
	u32 min_frame_size;

	/* OS defined structs */
	struct pci_dev *pdev;

	struct cdev char_dev;

	/* structs defined in e1000_hw.h */
	struct e1000_hw hw;

	/* user-dma specific variables */
	struct ei210_user_page	*userpages;
};

/* wrapper around a pointer to a socket buffer,
 * so a DMA handle can be stored along with the buffer
 */

struct ei210_tx_buffer {
	union e1000_adv_tx_desc *next_to_watch;
	DEFINE_DMA_UNMAP_ADDR(dma);
	DEFINE_DMA_UNMAP_LEN(len);
};

struct ei210_rx_buffer {
	dma_addr_t dma;
	struct page *page;
};

enum ei210_boards {
	board_82575,
};

static inline struct device *pci_dev_to_dev(struct pci_dev *pdev)
{
	return &pdev->dev;
}

#define ei210_write_reg(a, reg, value) ( \
	writel((value), ((a)->hw_addr + (reg))))

#define ei210_read_reg(a, reg) (readl((a)->hw_addr + (reg)))

static inline s32 ei210_reset_phy(struct e1000_hw *hw)
{
	if (hw->phy.ops.reset)
		return hw->phy.ops.reset(hw);

	return 0;
}

static inline s32 ei210_read_phy_reg(struct e1000_hw *hw, u32 offset, u16 *data)
{
	if (hw->phy.ops.read_reg)
		return hw->phy.ops.read_reg(hw, offset, data);

	return 0;
}

static inline s32 ei210_write_phy_reg(struct e1000_hw *hw, u32 offset, u16 data)
{
	if (hw->phy.ops.write_reg)
		return hw->phy.ops.write_reg(hw, offset, data);

	return 0;
}

static inline s32 ei210_get_phy_info(struct e1000_hw *hw)
{
	if (hw->phy.ops.get_phy_info)
		return hw->phy.ops.get_phy_info(hw);

	return 0;
}

#endif /* _NIECATI210_H_ */
